# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import *


def evaluate_mu(eq, vars):
    "Calculate mu, allowing that it is provided as a function"
    mu = vars['params'][eq.prefix + '.mu']
    if callable(mu):
        mu = mu(T=vars['T'], c=vars['c'][eq.prefix], E=None)
    vars['mu'][eq.prefix] = mu
    return mu


def species_v_D_charged_from_params(eq, vars):
    """
    Default implementation of transported species:
    - assumes constant mobility, takes from parameters
    - assumes diffusion described by Einstein's relation
    Note that mobility is saved for other calculations (such as recombination).
    """
    mu = evaluate_mu(eq, vars)
    if isscalar(mu):
        facemu = mu
    else:
        facemu = eq.mesh.faceaverage(mu)
    v = eq.v(eq.mesh.faceaverage(mu), vars['E'])
    D = eq.mesh.faceaverage(eq.D(mu, vars['Vt']))
    return v, D


def species_v_D_FrenkelPoole(eq, vars):
    mu0 = vars['params'][eq.prefix + '.mu0']
    gamma = vars['params'][eq.prefix + '.gamma']
    v = gamma * sqrt(vars['Ecellm'] + 1e-10)
    v_max = np.log(1e10)
    v_min = -v_max
    v = where(v < v_max, v, v_max)
    v = where(v > v_min, v, v_min)
    mu = mu0 * exp(v)
    vars['mu'][eq.prefix] = mu
    v = eq.v(eq.mesh.faceaverage(mu), vars['E'])
    D = eq.mesh.faceaverage(eq.D(mu, vars['Vt']))
    return v, D


def species_v_D_not_transported(eq, vars):
    """
    Not transported species
    """
    return None


def simple_doping(vars):
    params = vars['params']
    return (params['Nd'] - params['Na']) * scipy.constants.elementary_charge


def _check_samemesh(equations):
    for eq in equations[1:]:
        assert eq.mesh is equations[
            0].mesh, 'different mesheshes not supported'


def _check_samevars(eqvars, equations):
    vars = eqvars[id(equations[0])]
    for eq in equations[1:]:
        assert eqvars[id(eq)] is vars, 'equations in different models'


def species_v_D_limited_upwind(eq, vars):
    v, D = species_v_D_charged_from_params(eq, vars)
    c = where(v > 0,
              vars['c'][eq.prefix][eq.mesh.faces['+']],
              vars['c'][eq.prefix][eq.mesh.faces['-']])
    v = v * (1. - c / vars['params'][eq.prefix + '.N0'])
    return v, D



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import model
from oedes.ad import sparsesum
from oedes.fvm import Poisson, Transport
import numpy as np
import scipy.sparse
import itertools
import warnings
from oedes.util import TODOWarning
import logging
import scipy.sparse.csgraph

class PrefixDict:
    def __init__(self, d, prefix):
        if isinstance(d, PrefixDict):
            self.d = d.d
            self.prefix = d.prefix + '.' + prefix
        else:
            self.d = d
            self.prefix = prefix

    def _addprefix(self, k):
        if self.prefix is None:
            return k
        if k.startswith('__'):
            return k
        return self.prefix + '.' + k

    def __getitem__(self, k):
        k = self._addprefix(k)
        return self.d[k]

    def __setitem__(self, k, v):
        k = self._addprefix(k)
        return self.d.__setitem__(k, v)

    def update(self, d):
        for k in d:
            self[k] = d[k]

    def __contains__(self, k):
        return self.d.__contains__(self._addprefix(k))


def with_prefix(d, prefix):
    if d is None:
        return None
    return PrefixDict(d, prefix)


class vars_dict():

    def __init__(self, topvars, modelvars):
        self.topvars = topvars
        self.modelvars = modelvars

    def __getitem__(self, k):
        if k in self.modelvars:
            return self.modelvars[k]
        else:
            return self.topvars[k]

    def __setitem__(self, k, v):
        assert k not in self.modelvars
        self.modelvars[k] = v

    def update(self, d):
        for k in d:
            self[k] = d[k]

    def __contains__(self, k):
        return k in self.topvars or k in self.modelvars


logger = logging.getLogger('oedes.models.convergence')


class ComposableModel(model):
    name = None

    def __init__(self):
        model.__init__(self)
        self.after_load = []
        self.after_generate = []
        self.name = None

    @property
    def all_equations(self):
        "Return all equation in this model (and submodels)"
        raise NotImplementedError()

    def init_equations(self):
        for f in self.after_load:
            f.setUp(self)
        for f in self.after_generate:
            f.setUp(self)

    def load_equations(self, vars):
        raise NotImplementedError()

    def generate_residuals_equations(self, vars):
        raise NotImplementedError()

    def scaling(self, params):
        raise NotImplementedError('this is currently useless')

    def converged(self, *args):
        raise NotImplementedError('should not be called')

    def findeq(self, name):
        return dict((eq.prefix, eq) for eq in self.equations)[name]

    def _findeqs(self, test, prefix, return_names):
        for eq in self.equations:
            if test(eq):
                if return_names:
                    yield (prefix + eq.prefix, eq)
                else:
                    yield eq

    def findeqs(self, tests, return_names=False):
        return self._findeqs(tests, '', return_names=return_names)


class CompositeModel(ComposableModel):
    sub = None
    name = None
    after_load = None
    after_generate = None

    def __init__(self):
        self.sub = []
        self.name = None
        self.after_load = []
        self.after_generate = []

    @property
    def all_equations(self):
        return itertools.chain(
            self.equations, *(s.all_equations for s in self.sub))

    def create_indexes(self):
        i = 0
        for eq in self.all_equations:
            assert eq.idx is None, 'equation already has indices in unknown vector'
            eq.idx = np.arange(i, i + eq.mesh.ncells)
            i += eq.mesh.ncells
        self.X = np.zeros(i)

    def create_boundaries(self):
        i = [np.zeros(0, dtype=np.int)]
        j = [np.zeros(0, dtype=np.int)]
        for eq in self.all_equations:
            fdof = np.ones(len(eq.mesh.boundary.idx), dtype=np.bool)
            for bc in eq.bc:
                bc.setUp(eq)
                fdof[bc.bidx] = False
                valid = bc.conservation_to >= 0
                i.append(eq.idx[eq.mesh.boundary.idx[bc.bidx]][valid])
                j.append(bc.conservation_to[valid])
            eq.bc_free_dof = np.arange(len(eq.mesh.boundary.idx))[fdof]
        i = np.hstack(i)
        j = np.hstack(j)
        # optimize: should only create vector for boundary items
        g = scipy.sparse.csr_matrix(
            (np.ones_like(i), (i, j)), shape=(len(self.X),) * 2)
        nlabels, labels = scipy.sparse.csgraph.connected_components(
            g, directed=False, return_labels=True)
        self.bc_label_volume = sparsesum(nlabels, ((labels[eq.idx], eq.mesh.cells[
                                         'volume']) for eq in self.all_equations))
        self.bc_labels = labels
        for eq in self.all_equations:
            eq.boundary_labels = self.bc_labels[eq.idx[eq.mesh.boundary.idx]]

    def setUp(self):
        assert self.X is None, 'already set-up'
        self.create_indexes()
        for s in self.sub:
            s.init_equations()
        self.init_equations()
        self.create_boundaries()
        self.transientvar = self.X.copy()
        for eq in self.all_equations:
            self.transientvar[eq.idx] = eq.transientvar

    def load(self, time, x, xt, params, full_output, toplevel, solver):
        subvars = dict()
        modelvars = dict(time=time, x=x, xt=xt, solver=solver, model=self,
                         params=params, full_output=full_output, subvars=subvars)
        vars = vars_dict(toplevel, modelvars)
        # Set eqvars
        toplevel['modelvars'][id(self)] = vars
        eqvars = toplevel['eqvars']
        for eq in self.all_equations:
            eqvars[id(eq)] = modelvars
        # Load submodels
        for s in self.sub:
            assert s.name not in subvars, 'duplicate submodel name %s' % s.name
            subvars[s.name] = s.load(time, x, xt, params=with_prefix(
                params, s.name), full_output=with_prefix(full_output, s.name), toplevel=toplevel, solver=solver)
        # Load and evaluable variables equations in this model
        self.load_equations(vars)
        # Do additional calculations, which depend on
        for f in self.after_load:
            f.run(self, vars)
        return vars

    def generate_residuals(self, vars):
        # Return residuals, firslty submodels
        for s in self.sub:
            for res in s.generate_residuals(vars['subvars'][s.name]):
                yield res
        for res in self.generate_residuals_equations(vars):
            yield res
        for f in self.after_generate:
            f.run(self, vars)
        # return
        # itertools.chain(itertools.chain(s.generate_residuals(vars['subvars'][s.name])
        # for s in self.sub),self.generate_residuals(vars))

    def evaluate(self, time, x, xt, params, full_output=None, solver=None):
        vars = self.load(time, x, xt, params, full_output=full_output, toplevel=dict(
            bc_conservation=[], eqvars=dict(), modelvars=dict()), solver=solver)
        return itertools.chain(self.generate_residuals(
            vars), self.generate_bc_conservation(vars))

    def generate_bc_conservation(self, loc):
        bc_label_conservation = sparsesum(
            len(self.bc_labels), loc['bc_conservation'])
        for eq in self.all_equations:
            i = eq.idx[eq.mesh.boundary.idx[eq.bc_free_dof]]
            j = self.bc_labels[i]
            yield i, bc_label_conservation[j] * (1. / self.bc_label_volume[j])

    @property
    def equations(self):
        return []

    def generate_residuals_equations(self, vars):
        return []

    def load_equations(self, vars):
        pass

    def update_equations(self, x, params):
        pass

    def converged_equations(self, F, x, dx, params, report):
        return all([eq.testConverged(F, x, dx, params, report)
                    for eq in self.equations])

    # Things to be corrected!
    def scaling(self, params):
        # warnings.warn(
        #    'scaling  : should be corrected, nested, does not distinguish types of equations (BCs vs conservation)', TODOWarning)
        xscaling = np.ones_like(self.X)
        fscaling = np.ones_like(self.X)
        lunit = 1e-9
        tunit = 1e-12
        for eq in self.all_equations:
            if isinstance(eq, Poisson):
                fscaling[eq.idx] = lunit
            elif isinstance(eq, Transport):
                xscaling[eq.idx] = lunit**-3
                fscaling[eq.idx] = lunit * tunit
        return xscaling, fscaling

    def update(self, x, params):
        self.update_equations(x, params)
        for s in self.sub:
            sparams = with_prefix(params, s.name)
            s.update(x, sparams)

    def _converged(self, F, x, dx, params, report):
        return all([self.converged_equations(F, x, dx, params, report)] +
                   [sub._converged(F, x, dx, PrefixDict(params, sub.name), PrefixDict(report, sub.name)) for sub in self.sub])

    def converged(self, residuals, x, dx, params, report):
        c = self._converged(residuals, x, dx, params, report)
        for k in sorted(report.keys()):
            v = report[k]
            logger.debug('-------- %s converged=%r --------' %
                         (k, v['converged']))
            logger.debug('atol=%r, max(|F|)=%e (noncon.: %e)' %
                         (v['atol'], v['maxabs'], v['maxabs_nc']))
            logger.debug('rtol=%r, max(|x/dx|)=%e (noncon.: %e)' %
                         (v['rtol'], v['maxrel'], v['maxrel_nc']))
        return c

    def findeq(self, name):
        parts = name.split('.', 1)
        if len(parts) > 1:
            return dict((s.name, s)
                        for s in self.sub)[parts[0]].findeq(parts[1])
        else:
            return ComposableModel.findeq(self, name)

    def _findeqs(self, tests, prefix, return_names):
        return itertools.chain(ComposableModel._findeqs(self, tests, prefix, return_names),
                               *(s._findeqs(tests, prefix + s.name + '.', return_names) for s in self.sub))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import model
from oedes.ad import *
from oedes.fvm import BoundaryEquation, ElementwiseConvergenceTest
from oedes.fvm.poisson import Poisson
from oedes.fvm.transport import Transport, TransportCharged
from oedes.util import TODOWarning
from .compose import CompositeModel
import numpy as np
import scipy.constants
import itertools
import collections
import warnings


class BaseModel(CompositeModel):
    """
    Basic device model:
    Supports arbitrary number of species (transported or not) coupled to Poisson's equation, with any form of
    boundary conditions and drift and diffusion velocities
    """

    poisson = None
    species = None
    species_dos = None
    species_v_D = None
    sources = None
    other_equations = None
    additional_charge = None
    ordering = None
    after_load = None

    def __init__(self, *args, **kwargs):
        CompositeModel.__init__(self)
        self.species = []
        self.sources = []
        self.species_v_D = dict()
        self.species_dos = dict()
        self.other_equations = []
        self.additional_charge = []
        self.ordering = 'cell'

    def init_equations(self):
        assert isinstance(self.poisson, Poisson)
        assert len(set([eq.prefix for eq in self.all_equations])) == len(
            list(self.all_equations)), "duplicate prefix of equation"
        if self.poisson.convergenceTest is None:
            self.poisson.convergenceTest = ElementwiseConvergenceTest(
                atol=1e-15, rtol=1e-7)
            for bc in self.poisson.bc:
                if bc.convergenceTest is None:
                    bc.convergenceTest = ElementwiseConvergenceTest(
                        atol=1e-15, rtol=1e-7)
        for eq in self.species:
            assert isinstance(eq, Transport)
            if eq.convergenceTest is None:
                eq.convergenceTest = ElementwiseConvergenceTest(
                    atol=5e7, rtol=1e-7)
            for bc in eq.bc:
                if bc.convergenceTest is None:
                    bc.convergenceTest = ElementwiseConvergenceTest(
                        atol=1., rtol=1e-7)
        for eq in self.all_equations:
            assert eq.mesh is self.poisson.mesh, "multiple meshes not supported"

        CompositeModel.init_equations(self)

    def create_indexes(self):
        assert self.ordering in ['cell', 'equation']
        if self.ordering == 'equation':
            i = 0
            for eq in self.all_equations:
                assert eq.idx is None, "already indexed"
                ndof = eq.mesh.ncells
                eq.idx = np.arange(i, i + ndof, dtype=np.int)
                i += ndof
            self.X = np.zeros(i)
        else:
            neq = len(list(self.all_equations))
            for i, eq in enumerate(self.all_equations):
                assert eq.idx is None, "already indexed"
                assert eq.mesh is self.poisson.mesh, "multiple meshes not supported"
                eq.idx = np.arange(0, eq.mesh.ncells) * neq + i
            self.X = np.zeros(neq * self.poisson.mesh.ncells)

    @property
    def equations(self):
        return itertools.chain(
            [self.poisson], self.species, self.other_equations)

    def update_equations(self, x, params):
        # warnings.warn(
        #    'clipping to N0 also for ions, if specied in parameters : this could be very wrong!', TODOWarning)
        for eq in self.species:
            xe = x[eq.idx]
            x[eq.idx] = np.where(xe < 0., 0., xe)
            # if eq.prefix+'.N0' in params:
            #    x[eq.idx]=np.clip(xe,0.,params[eq.prefix+'.N0'])
            # else:
            #    x[eq.idx]=np.clip(xe,0.,xe)

    def evaluate_bc(self, eq, v, fluxes, FdS_boundary, vars,
                    celltransient_boundary=0., cellsource_boundary=0.):
        bclist = eq.bc
        conservation = -FdS_boundary + eq.mesh.cells['volume'][
            eq.mesh.boundary.idx] * (celltransient_boundary - cellsource_boundary)
        vars['bc_conservation'].append((eq.boundary_labels, conservation))
        for bc in bclist:
            yield eq.idx[eq.mesh.boundary.idx[bc.bidx]], bc.residuals(eq, v, FdS_boundary, vars)

    def evaluate_poisson(self, vars):
        "Part of evaluate dealing with Poisson's equation"
        c, params, E, Et, full_output = vars['c'], vars[
            'params'], vars['E'], vars['Et'], vars['full_output']
        epsilon = vars['epsilon']
        species_charge = sum([eq.ze * c[eq.prefix] for eq in self.species])
        additional_charge = sum([f(vars) for f in self.additional_charge])
        total_charge_density = species_charge + additional_charge
        faceepsilon = self.poisson.mesh.faceaverage(epsilon)
        D = self.poisson.displacement(E, faceepsilon)
        Dt = self.poisson.displacement(Et, faceepsilon)
        if full_output is not None:
            full_output['poisson.total_charge_density'] = total_charge_density
            full_output['potential'] = vars['potential']
            full_output['E'] = E
            full_output['Et'] = Et
            full_output['D'] = D
            full_output['Dt'] = Dt
        yield self.poisson.residuals(self.poisson.mesh.internal, D, cellsource=total_charge_density)
        FdS_boundary = dot(self.poisson.mesh.boundary.fluxsum, D)
        vars['Jd_boundary'] = dot(self.poisson.mesh.boundary.fluxsum, Dt)
        for res in self.evaluate_bc(self.poisson, vars[
                                    'potential'], D, FdS_boundary, vars):
            yield res

    def evaluate_source_terms(self, vars):
        "Part of evaluate dealing with source terms"

        sources = dict([(eq.prefix, 0.) for eq in self.species])
        for src in self.sources:
            f = src.evaluate(vars)
            for p in src.plus:
                sources[p] = sources[p] + f
            for n in src.minus:
                sources[n] = sources[n] - f
        return sources

    def evaluate_species(self, vars):
        "Part of evaluate dealing with residuals of species"

        E, sources, out = vars['E'], vars['sources'], vars['full_output']
        want_output = out is not None
        if out is not None:
            J = sum(out['Dt'] * self.poisson.mesh.faces['dr']) / \
                sum(self.poisson.mesh.faces['dr'])
        for eq in self.species:
            c = vars['c'][eq.prefix]
            ct = vars['ct'][eq.prefix]
            v_D = vars['v_D'][eq.prefix]
            if want_output:
                out[eq.prefix + '.c'] = c
                out[eq.prefix + '.ct'] = ct
            if v_D is None:
                j = np.zeros(len(eq.mesh.faces))
            else:
                v, D = v_D
                f = eq.faceflux(c=c, v=v, D=D, full_output=want_output)
                if want_output:
                    j = f['flux']
                    out[eq.prefix + '.j'] = f['flux']
                    out[eq.prefix + '.jdrift'] = f['flux_v']
                    out[eq.prefix + '.jdiff'] = f['flux_D']
                    if eq.prefix in vars['Ef']:
                        out[eq.prefix + '.Ef'] = vars['Ef'][eq.prefix]
                    J = J + sum(j * eq.ze *
                                eq.mesh.faces['dr']) / sum(eq.mesh.faces['dr'])
                else:
                    j = f
            yield eq.residuals(eq.mesh.internal, j, celltransient=ct, cellsource=sources[eq.prefix])
            FdS = dot(eq.mesh.boundary.fluxsum, j) + \
                sparsesum(len(eq.mesh.boundary.cells),
                          vars['boundary_FdS'][eq.prefix])
            vars['J_boundary'][eq.prefix] = FdS * eq.ze
            boundary_sources = sparsesum(len(eq.mesh.boundary.cells), vars[
                                         'boundary_sources'][eq.prefix])

            for res in self.evaluate_bc(eq, c, j, FdS, vars, celltransient_boundary=ct[
                                        eq.mesh.boundary.idx], cellsource_boundary=getitem(sources[eq.prefix], eq.mesh.boundary.idx) + boundary_sources):
                yield res
        if want_output:
            out['J'] = J

    # The design of this routine tries to reduce the cost of automatic differentation:
    # 1) Parts of x, xt corresponding to each field are taken once and put in c,ct etc.
    #    They should not be taken again in calculation routines.
    # 2) All calculation routines get common loc dict, which allows to store once
    #    calculated variables for reuse.
    # 3) The order of evaluation is as follows:
    #    - Poisson's equation
    #    - Species velocities v and diffusion coefficients D
    #      This usually involves calculation of mobilities which are not used
    #      directly by flux conservation, but are normally stored for sources.
    #    - Sources
    #    - Conservation of species (using sources, v and D)
    def load_equations(self, loc):
        x, xt, params = loc['x'], loc['xt'], loc['params']
        T = params['T']
        Vt = params['T'] * scipy.constants.Boltzmann / \
            scipy.constants.elementary_charge
        potential = x[self.poisson.idx]
        E = self.poisson.E(potential)
        Et = self.poisson.E(xt[self.poisson.idx])
        epsilon = scipy.constants.epsilon_0 * params['epsilon_r']
        Ecellv = self.poisson.mesh.cellaveragev(E)
        Ecellm = self.poisson.mesh.magnitudev(Ecellv)
        c = dict([(eq.prefix, x[eq.idx]) for eq in self.species])
        ct = dict([(eq.prefix, xt[eq.idx]) for eq in self.species])
        loc.update(dict(T=T, Vt=Vt, potential=potential, E=E, Et=Et, epsilon=epsilon, Ecellv=Ecellv, Ecellm=Ecellm,
                        c=c, ct=ct, FdS_boundary={}, J_boundary={}, Ef={}, dos={}))
        for eq in self.species:
            if eq.prefix in self.species_dos:
                dos = self.species_dos[eq.prefix]
                Ef = dos.QuasiFermiLevel(eq, loc)
                loc['Ef'][eq.prefix] = Ef
                loc['dos'][eq.prefix] = dos
        loc['mu'] = dict()
        loc['v_D'] = dict([(eq.prefix, self.species_v_D[eq.prefix](eq, loc))
                           for eq in self.species if eq.prefix])
        loc['sources'] = self.evaluate_source_terms(loc)
        loc['boundary_FdS'] = dict([(eq.prefix, []) for eq in self.species])
        loc['boundary_sources'] = dict(
            [(eq.prefix, []) for eq in self.species])

    def generate_residuals_equations(self, loc):
        for res in self.evaluate_poisson(loc):
            yield res
        for res in self.evaluate_species(loc):
            yield res
        # for res in self.generate_bc_conservation(loc):
        #    yield res



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.ad import *


def test_sparsesum():
    idx, v = sparsesum(10, [(0, 1), (0, 1), (3, -1)], compress=True)
    assert (idx == np.asarray([0, 3])).all()
    assert (v == np.asarray([2., -1.])).all()

    x = forward.seed(np.linspace(0, 1, 11))
    idx, d = sparsesum(
        20, [(np.arange(11), x), (np.arange(11), x)], compress=True)
    assert (idx == np.arange(11)).all()
    assert (d.gradient.diagonal() == 2.).all()

    d = sparsesum(1, [(np.zeros(2, dtype=int), np.ones(2)),
                      (np.zeros(1, dtype=int), np.ones(1))])
    assert d == 3



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .dos import DOS
from oedes.ad import *
import scipy.constants
import oedes.functions.egdm as _egdm
import oedes.functions.gdos as _gdos
from oedes.functions.physics import ThermalVoltage
import numpy as np
import warnings


def egdm_params_simple(params, prefix, nsigma, a, mu0t):
    params[prefix + '.N0'] = _egdm.atoN0(a)
    params[
        prefix + '.sigma'] = _egdm.nsigma_inverse(nsigma, ThermalVoltage(params['T']))
    params[prefix + '.mu0'] = _egdm.mu0t_inverse(nsigma, mu0t)


class GaussianDOS(DOS):
    # it is allowed to turn off any enhancement factor
    use_g1 = True
    use_g2 = True
    use_g3 = True
    # to avoid numeric problems, c is clipped to more than c_eps
    c_eps = 1e-30
    c_max = 0.9999
    # custom limits are allowed
    g1_max_c = _egdm.g1_max_c
    g2_max_En = _egdm.g2_max_En
    g3_max_c = _egdm.g3_max_c
    # custom gdos impl is allowed, but must be compatible with
    # limits set here
    gdosImpl = _gdos.defaultImpl

    # To avoid numeric problems, nsigma is clipped to these values.
    # This can be useful in initial Newton iterations with
    # thermal coupling when temperature can be weird, and so nsigma.
    # However, final solution should not be affected by that.
    # Keep in mind that the model is for nsigma=3,4,5,6. nsigma<=1
    # will cause NaN in g1, and nsigma>>6 is problematic for Gauss-
    # Fermi integral.
    nsigma_min = 2.
    nsigma_max = 7.

    def __init__(self, impl=None):
        DOS.__init__(self)
        if impl is None:
            impl = _gdos.defaultImpl
        self.impl = impl

    def _load(self, eq, vars):
        params = vars['params']
        sigma = params[eq.prefix + '.sigma']
        Vt = vars['Vt']
        nsigma = _egdm.nsigma(sigma, Vt)
        nsigma = where(nsigma > self.nsigma_min, nsigma, self.nsigma_min)
        nsigma = where(nsigma < self.nsigma_max, nsigma, self.nsigma_max)
        if np.any(_egdm.nsigma(sigma, Vt) != nsigma):
            warnings.warn(
                'nsigma clipped to ensure numerical stability: probably not calculating what you want. This is OK in initial Newton iterations, but not in the final one')
        N0 = params[eq.prefix + '.N0']
        c = vars['c'][eq.prefix] / N0
        assert self.c_eps >= self.impl.I_min
        assert self.c_max <= self.impl.I_max
        c = where(c > self.c_eps, c, self.c_eps)
        c = where(c < self.c_max, c, self.c_max)
        return sigma, nsigma, N0, Vt, c

    def Ef(self, eq, vars):
        _, nsigma, N0, Vt, c = self._load(eq, vars)
        return -self.impl.b(np.sqrt(2.) * nsigma, c) * Vt

    def c(self, eq, Ef, vars):
        _, nsigma, N0, Vt, c = self._load(eq, vars)
        return N0 * self.impl.I(np.sqrt(2.) * nsigma, b=-Ef / Vt)

    def D(self, eq, c, Ef, vars):
        raise NotImplementedError()

    def v_D(self, eq, vars):
        sigma, nsigma, N0, Vt, c = self._load(eq, vars)
        params = vars['params']
        a = _egdm.N0toa(N0)
        mu_cell = _egdm.mu0t(nsigma, params[eq.prefix + '.mu0'])
        if self.use_g1:
            mu_cell = mu_cell * \
                _egdm.g1(nsigma, where(c < self.g1_max_c, c, self.g1_max_c))
        mu_face = eq.mesh.faceaverage(mu_cell)
        if self.use_g2:
            En = (vars['E'] * a) / sigma
            En2 = En**2
            En2_max = self.g2_max_En**2
            mu_face = mu_face * \
                _egdm.g2_En2(nsigma, where(En2 < En2_max, En2, En2_max))
        v = eq.v(mu_face, vars['E'])
        D_face = mu_face * Vt
        if self.use_g3:
            assert self.g3_max_c <= self.gdosImpl.I_max
            assert self.c_eps >= self.gdosImpl.I_min
            assert self.nsigma_max * np.sqrt(2.) <= self.gdosImpl.a_max
            D_face = D_face * \
                eq.mesh.faceaverage(_egdm.g3(nsigma, where(
                    c < self.g3_max_c, c, self.g3_max_c)))
        return eq.v(mu_face, vars['E']), D_face



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import matplotlib
import matplotlib.pylab as plt
import matplotlib.ticker
from collections import defaultdict
from oedes.fvm import mesh1d
import numpy as np
import itertools
from matplotlib.collections import LineCollection

__all__ = ['matplotlib', 'plt', 'forcontext', 'subplots', 'Settings']


class mesh1d_mpl(object):
    def __init__(self, fig, ax):
        self.fig = fig
        self.ax = ax
        self.cycler = itertools.cycle(matplotlib.rcParams['axes.prop_cycle'])

    def plot(self, items, face, *args, **kwargs):

        def map(item):
            mesh, value = item
            if not isinstance(mesh, mesh1d):
                raise ValueError('mesh type not supported')
            if face:
                X = mesh.faces['center']
            else:
                X = mesh.cells['center']
            return np.vstack((X, value)).transpose()
        if 'color' not in kwargs:
            kwargs['color'] = next(self.cycler)['color']
        self.ax.add_collection(LineCollection(
            [map(i) for i in items], *args, **kwargs))
        self.ax.autoscale()

    def settings(self, settings):

        for name in ['xlabel', 'ylabel', 'xlim', 'ylim', 'xscale', 'yscale']:
            if name in settings:
                getattr(self.ax, 'set_' + name)(settings[name])

        prefixes = {'n': 1e-9, r'\mu': 1e-6, 'u': 1e-6, 'm': 1e-3, None: 1.}

        def aunit(axobj, unit):
            try:
                pv = prefixes[unit]
            except BaseException:
                pv = float(unit)
            axobj.set_major_formatter(matplotlib.ticker.FuncFormatter(
                lambda x, pos: '{0:g}'.format(x / pv)))
            axobj.set_minor_formatter(matplotlib.ticker.NullFormatter())

        if 'xunit' in settings:
            aunit(self.ax.xaxis, settings['xunit'])
        if 'yunit' in settings:
            aunit(self.ax.yaxis, settings['yunit'])


import re


class Settings:
    """
    Plot presets, following naming
    yquantity[_yunit-if-not-basic]_xquantity[_xunit_if_not_basic][_[log_or_lin_y][log_or_lin_x]]
    """

    c_x_nm_loglin = dict(
        yscale='log',
        ylabel=r'$\mathrm{m^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_cm_x_nm_loglin = dict(
        yscale='log',
        yunit=1e6,
        ylabel=r'$\mathrm{cm^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_x_nm = dict(
        ylabel=r'$\mathrm{m^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_cm_x_nm = dict(
        yunit=1e6,
        ylabel=r'$\mathrm{cm^{-3}}$',
        xunit='n',
        xlabel='nm')
    potential_x_nm = dict(ylabel=r'V', xunit='n', xlabel='nm')
    E_x_nm = dict(ylabel=r'$\mathrm{V \ m^{-1}}$', xunit='n', xlabel='nm')
    Ef_x_nm = dict(ylabel=r'eV', xunit='n', xlabel='nm')

    j_V = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel='V')
    j_V_loglin = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='V',
        yscale='log')
    j_V_loglog = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='V',
        yscale='log',
        xscale='log')

    j_mAcm_V = dict(yunit=10, ylabel=r'$\mathrm{mA \ cm^{-2}}$', xlabel='V')

    j_t = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel='s')
    j_t_loglog = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='s',
        xscale='log',
        yscale='log')
    j_t_us = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel='$\mu s$', xunit='u')
    j_mAcm_t_us = dict(
        yunit=10,
        ylabel=r'$\mathrm{mA \ cm^{-2}}$',
        xlabel='$\mu s$',
        xunit='u')


class context_plotting:
    def __init__(self, ctx, fig=None, ax=None):
        self.out = ctx.output()
        self.model = ctx.model
        self.ctx = ctx
        self.fig = fig
        if ax is not None:
            self.ax = ax
        else:
            self.fig, self.ax = plt.subplots()
        self._impl = None

        self.default_settings_species = Settings.c_x_nm_loglin
        self.default_settings_potential = Settings.potential_x_nm
        self.default_settings_E = Settings.E_x_nm
        self.default_settings_Ef = Settings.Ef_x_nm

    def getimpl(self, mesh):
        if self._impl is None:
            if mesh is None or isinstance(mesh, mesh1d):
                c = mesh1d_mpl
            elif hasattr(mesh, 'mpl_class'):
                c = mesh.mpl_class
            else:
                raise ValueError(
                    'do not know mpl_class of %s' %
                    mesh.__class__)
            self._impl = c(self.fig, self.ax)
        return self._impl

    def select(self, expr):
        r = re.compile(expr)
        grouped = defaultdict(list)
        with_label = 'label' in r.groupindex
        for k in sorted(self.out.keys()):
            match = r.match(k)
            if match:
                model = self.model
                path = k.split('.')
                while path:
                    sub = dict([(s.name, s) for s in model.sub])
                    if path[0] not in sub:
                        break
                    model = sub[path.pop(0)]
                ent = (model.poisson.mesh, self.out[k])
                if not with_label:
                    yield k, [ent]
                else:
                    grouped[match.group('label')].append(ent)
        for k in sorted(grouped.keys()):
            yield k, grouped[k]

    def _plot(self, parts, face=False, settings={}, *args, **kwargs):
        if not parts:
            return
        impl = self.getimpl(parts[0][0])
        impl.plot(parts, *args, face=face, **kwargs)
        impl.settings(settings)

    def selector(self, expr, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = default_settings
        want_label = 'label' not in kwargs
        for legendkey, parts in self.select(expr):
            if want_label:
                kwargs['label'] = legendkey
            self._plot(parts, *args, **kwargs)

    def allspecies(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_species
        self.selector(r'(.*\.)?(?P<label>.*)\.c$', *args, **kwargs)

    def species(self, prefix, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_species
        self.selector(r'(.*\.)?%s\.c$' % re.escape(prefix), *args, **kwargs)

    def potential(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_potential
        self.selector(r'(.*\.)?(?P<label>potential)$', *args, **kwargs)

    def E(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_E
        self.selector(r'(.*\.)?(?P<label>E)$', *args, face=True, **kwargs)

    def Ef(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_Ef
        self.selector(r'(.*\.)?(?P<label>.*)\.Ef$', *args, **kwargs)

    def teval(self, *functions, **kwargs):
        settings = kwargs.pop('settings', None)
        results = self.ctx.teval(*functions)
        want_label = 'label' not in kwargs
        for i in range(1, len(results)):
            if want_label:
                kwargs['label'] = str(functions[i])
            self.ax.plot(results[0], results[i], **kwargs)
        self.ax.set_xlabel(str(functions[0]))
        if settings:
            self.getimpl(None).settings(settings)

    def apply(self, settings):
        self.impl.apply(settings)


def subplots(c, *args, **kwargs):
    fig, u = plt.subplots(*args, **kwargs)

    def wrap(x):
        if isinstance(x, tuple):
            return tuple(wrap(y) for y in x)
        elif isinstance(x, list):
            return list(wrap(y) for y in x)
        else:
            return context_plotting(c, fig, x)
    return wrap(u)


def forcontext(*args, **kwargs):
    return context_plotting(*args, **kwargs)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import ad
import numpy as np
import scipy.special
import scipy.integrate


class AbstractGaussFermiIntegralImpl(object):
    """
    Abstract implementation of Gauss-Fermi integral:
    I(a,b) = 1/sqrt(pi) \int_{-\infty}{+\infty} dx exp(-x^2)/(1.+exp(a*x+b))

    For simulations, three functions have to be provided in a consistent way:
    - forward calculation, provided by method I(a,b)
    - inverse calculation solving I(a,b)=I for b, provided by method b(a,I)
    - derivative calculation dI/db, provided by method dIdb(a,b)

    The last one is necessary because sparsegrad support is limited to one order
    of differentiation, therefore a function calculating first derivative has to
    be explicitly called to enable automatic calculation of second-order
    derivative.

    It is quite hard to give practical approximation of the integral which would
    work for all values. Therefore range of supported parameters values is
    also specified:
    - 0<=a<=a_max (for all calculations)
    - I_min<=I<=I_max (for calculation of b solving I(a,b)=I)
    """

    maxiter = 20
    a_max = None
    I_min = None
    I_max = None

    def nevaluate(self, a, b, need):
        """
        Perform numerical evaluation of needed quantities need with known
        parameters a,b. In need, 'I' denotes I(a,b), 'da' denotes dI/da,
        'db' denotes dI/db, 'd2ab' denotes d2I/(dadb) and 'd2bb' denotes
        d2I/db2.
        """
        raise NotImplementedError()

    def nsolve_b(self, a, i, b0=None, halley=True, tol=1e-14):
        """
        Solve equation I(a,b)=i for b with known a.
        """
        if b0 is None:
            # 1e-20 is added to avoid NaN, has no effect on calculation
            b = np.where(i < 0.1, -np.log(i) + 0.25 * a **
                         2, np.log(1.0 / (i + 1e-20) - 1.))
        else:
            b = b0
        need = ['I', 'db']
        if halley:
            need.append('d2bb')
        for itno in range(self.maxiter):
            values = self.nevaluate(a, b, need=need)
            I = values['I']
            df = values['db']
            if halley:
                d2f = values['d2bb']
            else:
                d2f = 0.
            f = I - i
            if np.amax(np.abs(f) / i) < tol:
                return b
            delta = 2. * f * df / (2. * df * df - f * d2f)
            if not np.isfinite(delta).all():
                raise RuntimeError('numeric error')
            b -= delta
        raise RuntimeError('iteration limit exceeded')

    def I(self, a, b):
        def _I(a, b):
            return self.nevaluate(a, b, need=['I'])['I']

        def _I_deriv(args, value):
            a, b = args
            values = self.nevaluate(a, b, need=['da', 'db'])
            yield lambda: values['da']
            yield lambda: values['db']
        return ad.custom_function(_I, _I_deriv)(a, b)

    def dIdb(self, a, b):
        def _dIdb(a, b):
            return self.nevaluate(a, b, need=['db'])['db']

        def _dIdb_deriv(args, value):
            a, b = args
            values = self.nevaluate(a, b, need=['d2ab', 'd2bb'])
            yield lambda: values['d2ab']
            yield lambda: values['d2bb']
        return ad.custom_function(_dIdb, _dIdb_deriv)(a, b)

    def b(self, a, i, b0=None):
        def _b(a, i):
            return self.nsolve_b(a, i, b0=b0, halley=True)

        def _b_deriv(args, value):
            a, i = args
            b = value
            values = self.nevaluate(a, b, need=['da', 'db'])
            dIda, dIdb = values['da'], values['db']
            dbdI = 1. / dIdb
            yield lambda: -dbdI * dIda
            yield lambda: dbdI
        return ad.custom_function(_b, _b_deriv)(a, i)

    def __call__(self, a, b):
        return self.I(a, b)


class GaussFermiIntegralH(AbstractGaussFermiIntegralImpl):
    """GaussFermiIntegral implemented using Gauss-Hermite quadrature"""

    def __init__(self, n, a_max):
        AbstractGaussFermiIntegralImpl.__init__(self)
        self.maxiter = 20
        self.a_max = a_max
        self.I_min = 1e-30
        self.I_max = 0.9999
        _x, _w = scipy.special.h_roots(n)
        self.X_W = list(zip(_x, _w / np.sqrt(np.pi)))

    def nevaluate(self, a, b, need):
        I, da, db, d2ab, d2bb = None, None, None, None, None
        if 'I' in need:
            I = 0.
        if 'da' in need:
            da = 0.
        if 'db' in need:
            db = 0.
        if 'd2ab' in need:
            d2ab = 0.
        if 'd2bb' in need:
            d2bb = 0.
        need_first = da is not None or db is not None
        need_second = d2ab is not None or d2bb is not None
        need_deriv = need_first or need_second
        for x, w in self.X_W:
            axpb = a * x + b
            u = np.exp(axpb)
            f = 1. / (1. + u)
            if I is not None:
                I += w * f
            if need_deriv:
                u_f = np.where(axpb < 40., u * f, 1.)
                df = -f * u_f
                w_df = w * df
                if need_first:
                    if da is not None:
                        da += w_df * x
                    if db is not None:
                        db += w_df
                if need_second:
                    w_ddf = w_df * (1. - 2. * u_f)
                    if d2ab is not None:
                        d2ab += w_ddf * x
                    if d2bb is not None:
                        d2bb += w_ddf
                    # nb: d2aa=w_ddf*x*x
        loc = locals()
        return dict((k, loc[k]) for k in need)


class GaussFermiIntegralReference(AbstractGaussFermiIntegralImpl):
    """
    Reference implementation for checking accuracy, which uses adaptive
    quadrature.
    Only calculation of I is supported.
    """

    def nevaluate(self, a, b, need):
        if not need:
            return dict()
        if list([v for v in need if v != 'I']):
            raise NotImplementedError()

        def _f(x):
            return np.exp(-x * x) / (1 + np.exp(a * x + b))
        v = scipy.integrate.quad(_f, -28, 28,
                                 epsrel=1e-12, epsabs=1e-40)
        return dict(I=v[0] / np.sqrt(np.pi))


def diffusion_enhancement(nsigma_as_a, c, impl, full_output=False):
    b = impl.b(nsigma_as_a, c)
    factor = -c / impl.dIdb(nsigma_as_a, b)
    if full_output:
        return dict(factor=factor, b=b)
    else:
        return factor


defaultImpl = GaussFermiIntegralH(96, 7. * np.sqrt(2.))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from sparsegrad import *
import numpy as np
import scipy.sparse

nvalue = forward.nvalue


def isscalar(x):
    return forward.isscalar(forward.nvalue(x))


def custom_function(f, df):
    return wrapped_func(func.ufunc(func=f, deriv=df))


def sparsesum(n, terms, compress=False, unique=False):
    """
    Sum sparse vectors, where each term is given as (index,values)
    Resulting length n should be known in advance
    If compress, return as a spearse vector idx,sum
    If unique, only one term can contribute to each result
    """
    # s=np.zeros(n)
    sdata = [np.zeros(0, dtype=np.double)]
    scol = [np.zeros(0, dtype=np.int)]
    data = [np.zeros(0, dtype=np.double)]
    row = [np.zeros(0, dtype=np.int)]
    col = [np.zeros(0, dtype=np.int)]
    M = None
    used = np.zeros(n, dtype=np.bool)
    for idx, f in terms:
        v = np.asarray(nvalue(f))
        # result_dtype=np.result_type(s.dtype,v.dtype)
        # if s.dtype is not result_dtype:
        #    s=np.asarray(s,dtype=s.dtype)
        sdata.append(v)
        scol.append(idx)
        # s[idx]+=v
        if unique and np.any(used[idx]):
            raise RuntimeError('duplicate values')
        used[idx] = True
        if isinstance(f, expr_base):
            M = f.M
            fx = f.gradient.tocsr().tocoo()
            gw = fx.shape[1]
            assert fx.shape[0] == len(idx)
            row.append(idx[fx.row])
            col.append(fx.col)
            data.append(fx.data)
    sdata = np.hstack(sdata)
    scol = np.hstack(scol)
    s = np.asarray(scipy.sparse.csr_matrix(
        (sdata, (np.zeros_like(scol), scol)), shape=(1, n)).todense()).flatten()
    if M is None:
        result = s
    else:
        data = np.hstack(data)
        col = np.hstack(col)
        row = np.hstack(row)
        d = scipy.sparse.csr_matrix((data, (row, col)), shape=(s.shape[0], gw))
        result = forward.value(value=s, dvalue=d, M=M)
    if compress:
        return np.arange(n)[used], result[used]
    else:
        return result


def getitem(x, idx):
    if isscalar(x):
        return x
    else:
        return x[idx]



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import solver
from . import util
from .fvm import Poisson
from .models import RestrictedModel

from collections import namedtuple
import numpy as np
import bisect

contextstate = namedtuple('contextstate', ['x', 'xt', 'time', 'params'])


def _bsearch(a, cmp):
    i, j, k = 0, len(a), -1
    while i < j:
        k = (i + j) // 2
        c = cmp(a[k])
        if c == 0:
            break
        elif c < 0:
            i = k + 1
        else:
            j = k
    if cmp(a[k]) > 0:
        return k - 1
    return k


_progress_msg = []


class progressinfo:

    def __init__(self, msg):
        self.msg = msg

    def __enter__(self):
        _progress_msg.append(self.msg)

    def __exit__(self, type, value, traceback):
        _progress_msg.pop()


class context:

    def __init__(self, model_or_context, **kwargs):
        if isinstance(model_or_context, context):
            ctx = model_or_context
            self.model = ctx.model
            self.x = ctx.x
            self.xt = ctx.xt
            self.time = ctx.time
            self.params = ctx.params
            self._output = ctx._output
            self.trajectory = ctx.trajectory
            self._tsdict = dict(ctx._tsdict)
        else:
            model = model_or_context
            self.model = model
            self.x = np.zeros_like(model.X)
            self.xt = np.zeros_like(model.X)
            self.time = 0.
            self.params = {}
            self._output = None
            self.trajectory = []
            self._tsdict = {}
        self.set(**kwargs)

    def progress(self, ts):
        util.progress('%s time=%s' % (' '.join(_progress_msg), ts.time))

    def state(self):
        return context(self.model, x=self.x, xt=self.xt,
                       time=self.time, params=self.params)

    def restore(self, state):
        assert self.model is state.model
        self.set(time=state.time, x=state.x, xt=state.xt, params=state.params)

    def set(self, time=None, x=None, xt=None, params=None):
        if time is not None:
            self.time = time
        if x is not None:
            self.x = x
        if xt is not None:
            self.xt = xt
        if params is not None:
            self.params = params
        self._output = None
        self._acsolver = None

    def bdf1adapt(self, params, t1, dt, *args, **kwargs):
        params = dict(params)
        self.trajectory.append(contextstate(
            time=self.time, x=self.x, xt=self.xt, params=params))
        for time, x, xt, out in solver.bdf1adapt(
                self.model, self.x, params, self.time, t1, dt, *args, **kwargs):
            self.set(time=time, x=x, xt=xt, params=params)
            self.trajectory.append(self.state())
            yield self.state()

    def transientsolve(self, params, timesteps, *args, **kwargs):
        params = dict(params)
        self.trajectory.append(contextstate(
            time=self.time, x=self.x, xt=self.xt, params=params))
        for time, x, xt, out in solver.transientsolve(
                self.model, self.x, params, timesteps, *args, **kwargs):
            self.set(time=time, x=x, xt=xt, params=params)
            self.trajectory.append(self.state())
            yield self.state()

    def transient(self, *args, **kwargs):
        for ts in self.bdf1adapt(*args, **kwargs):
            self.progress(ts)

    def _psolve(self, params, xguess):
        eqs = [
            name for name,
            eq in self.model.findeqs(
                lambda eq:isinstance(
                    eq,
                    Poisson),
                return_names=True)]
        if not eqs:
            return xguess
        r = RestrictedModel(self.model, eqs, xguess, 0. * xguess)
        phi = solver.solve(r, r.X, params)
        return r.x_to_model(phi)

    def _solve(self, xguess, params, pseudo_tmax=1e6, pseudo_dt0=1e-9,
               pseudo_mindt=1e-15, pseudo_maxsteps=200, use_poisson_guess=True, **kwargs):
        if use_poisson_guess:
            xguess = self._psolve(params, xguess=xguess)
        try:
            return solver.solve(self.model, xguess, params, **kwargs)
        except solver.SolverError:
            for t, x, xt, out in solver.bdf1adapt(
                    self.model, xguess, params, 0., pseudo_tmax, pseudo_dt0, mindt=pseudo_mindt, maxsteps=pseudo_maxsteps, **kwargs):
                pass
            return solver.solve(self.model, x, params, **kwargs)

    def sweep(self, parameters, **kwargs):
        xguess = self.x
        for params in parameters:
            x = self._solve(xguess, params, **kwargs)
            xguess = x
            self.set(time=0., x=x, params=params)
            self.trajectory.append(self.state())
            yield self.state()

    def solve(self, params, xguess=None, **kwargs):
        params = dict(params)
        if xguess is None:
            xguess = self.x
        x = self._solve(xguess, params, **kwargs)
        self.set(time=0., x=x, xt=np.zeros_like(x), params=params)
        self.trajectory.append(self.state())

    def output(self):
        if self._output is None:
            self._output = self.model.output(
                self.time, self.x, self.xt, self.params)
        return self._output

    @property
    def pylab(self):
        import oedes.mpl as mpl
        return mpl.forcontext(self)

    def mpl(self, fig, ax):
        import oedes.mpl as mpl
        return mpl.forcontext(self, fig, ax)

    def attime(self, time):
        i = 0
        if not self.trajectory:
            raise ValueError
        i = _bsearch(self.trajectory, lambda ts: ts.time - time)
        if i < 0:
            raise ValueError
        ts0 = self.trajectory[i]
        assert ts0.time <= time
        if ts0.time < time:
            if (i + 1) >= len(self.trajectory):
                raise ValueError()
            ts1 = self.trajectory[i + 1]
            assert ts1.time > time
            assert ts0.params is ts1.params
            k = (time - ts0.time) / (ts1.time - ts0.time)
            ts = contextstate(time=time, x=ts0.x + (ts1.x - ts0.x) * k,
                              xt=ts0.xt + k * (ts1.xt - ts0.xt), params=ts1.params)
        else:
            ts = ts0
        return context(self.model, time=ts.time, x=ts.x,
                       xt=ts.xt, params=ts.params)

    def timesteps(self):
        for ts in self.trajectory:
            if id(ts) not in self._tsdict:
                self._tsdict[id(ts)] = context(
                    self.model, time=ts.time, x=ts.x, xt=ts.xt, params=ts.params)
            yield self._tsdict[id(ts)]

    def eval_function(self, f):
        if callable(f):
            return f(self)
        elif f in self.output():
            return self.output()[f]
        elif f in self.params:
            return self.params[f]
        elif f == 'time':
            return self.time
        else:
            raise RuntimeError('cannot find %s' % f)

    def teval_function(self, f):
        return np.asarray([ts.eval_function(f) for ts in self.timesteps()])

    def eval(self, *functions):
        return [self.eval_function(f) for f in functions]

    def teval(self, *functions):
        return [self.teval_function(f) for f in functions]

    def acsolver(self, input_param, output_function_or_name, **kwargs):
        if callable(output_function_or_name):
            output = output_function_or_name
        else:
            assert isinstance(output_function_or_name, str)

            def output(model, time, x, xt, p, full_output):
                return full_output[output_function_or_name]
        return solver.ACSolver(self.model, self.time, self.x,
                               self.xt, self.params, input_param, output, **kwargs)


def setparam(params, pname, values):
    for x in values:
        p = dict(params)
        p[pname] = x
        yield p
